/* Copyright (c) 2004 Joshua Baker
   See the file COPYING for copying permission.
*/

#include <setjmp.h>
#include <string.h>
#include "xaint.h"

static XaintBool __push_record_name(XaintStream *xaintst, XaintChar *name);
static void __free_record_name(XaintStream *xaintst);
static void __pop_record_name(XaintStream *xaintst);

typedef struct __recordName_type __recordName;
struct __recordName_type
{
  XaintChar *recordName;
  __recordName *next;
};

struct XaintStream_type
{
  XaintUnknown userData;
  XaintUnknown privData;
  XaintParser *xaintpr;
  XaintChar *key;
  XaintChar *ws;
  XaintBool storeRecordNames;
  XaintUnknown recordNames;
  XaintStreamPairEvent pairEvent;
  XaintStreamWhitespaceEvent wsEvent;
  XaintStreamPragmaEvent pragmaEvent;
  XaintStreamCommentEvent commentEvent;
  XaintStreamDataRequest dataRequest;
  XaintUnknown ClassInstance;
  XaintStreamOnDelete OnDelete;
};



XaintParser *
XaintStream_INTERNAL_get_xaintpr(XaintStream *xaintst)
{
  return xaintst->xaintpr;
}

XaintStream 
*XaintStream_new()
{
  XaintStream *xaintst;
  xaintst = (struct XaintStream_type*)Xaint_malloc(sizeof(struct XaintStream_type));
  if (!xaintst) return 0;
  memset(xaintst, 0, sizeof(struct XaintStream_type));
  xaintst->xaintpr = XaintParser_new();
  if (!xaintst->xaintpr) 
  {
    XaintStream_delete(xaintst);
    return 0;
  }
  XaintParser_setUserData(xaintst->xaintpr, xaintst);
  return xaintst;
}

void 
XaintStream_delete(XaintStream *xaintst)
{
  if (!xaintst) return;
  if (xaintst->OnDelete) xaintst->OnDelete(xaintst);
  
  if (xaintst->key) 
  { 
    Xaint_free(xaintst->key); 
    xaintst->key = 0; 
  }
  if (xaintst->ws) 
  { 
    Xaint_free(xaintst->ws); 
    xaintst->ws = 0; 
  }
  __free_record_name(xaintst);
  if (xaintst->xaintpr) XaintParser_delete(xaintst->xaintpr);
  
  
  Xaint_free(xaintst);
}


void 
XaintStream_setPairEvent(XaintStream *xaintst, 
                         XaintStreamPairEvent event)
{
  if (!xaintst) return;
  xaintst->pairEvent = event;
}

XaintStreamPairEvent
XaintStream_getPairEvent(XaintStream *xaintst)
{
  if (!xaintst) return 0;
  return xaintst->pairEvent;
}

void 
XaintStream_setWhitespaceEvent(XaintStream *xaintst, 
                               XaintStreamWhitespaceEvent event)
{
  if (!xaintst) return;
  xaintst->wsEvent = event;
}

XaintStreamWhitespaceEvent
XaintStream_getWhitespaceEvent(XaintStream *xaintst)
{
  if (!xaintst) return 0;
  return xaintst->wsEvent;
}

void 
XaintStream_setPragmaEvent(XaintStream *xaintst, 
                           XaintStreamPragmaEvent event)
{
  if (!xaintst) return;
  xaintst->pragmaEvent = event;
}

XaintStreamPragmaEvent 
XaintStream_getPragmaEvent(XaintStream *xaintst)
{
  if (!xaintst) return 0;
  return xaintst->pragmaEvent;
}

void 
XaintStream_setCommentEvent(XaintStream *xaintst, 
                            XaintStreamCommentEvent event)
{
  if (!xaintst) return;
  xaintst->commentEvent = event;
}

XaintStreamCommentEvent
XaintStream_getCommentEvent(XaintStream *xaintst)
{
  if (!xaintst) return 0;
  return xaintst->commentEvent;
}


void 
XaintStream_setDataRequest(XaintStream *xaintst, 
                           XaintStreamDataRequest dataRequest)
{
  if (!xaintst) return;
  xaintst->dataRequest = dataRequest;
}

XaintStreamDataRequest
XaintStream_getDataRequest(XaintStream *xaintst)
{
  if (!xaintst) return 0;
  return xaintst->dataRequest;
}

void 
XaintStream_setUserData(XaintStream *xaintst, XaintUnknown userData)
{
  if (!xaintst) return;
  xaintst->userData = userData;
}


XaintUnknown 
XaintStream_getUserData(XaintStream *xaintst)
{
  if (!xaintst) return 0;
  return xaintst->userData;
}

XaintErr 
XaintStream_getLastError(XaintStream *xaintst)
{
  if (!xaintst) return XAINT_ERR_NIL;
  return XaintParser_getLastError(xaintst->xaintpr);
}


XaintNum
XaintStream_getRowNum(XaintStream *xaintst)
{
  if (!xaintst) return 0;
  return XaintParser_getRowNum(xaintst->xaintpr);
}

XaintNum
XaintStream_getColNum(XaintStream *xaintst)
{
  if (!xaintst) return 0;
  return XaintParser_getColNum(xaintst->xaintpr);
}

void 
XaintStream_setStoreRecordNames(XaintStream *xaintst, XaintBool storeRecordNames)
{
  if (!xaintst) return;
  xaintst->storeRecordNames = storeRecordNames;
}

XaintBool 
XaintStream_getStoreRecordNames(XaintStream *xaintst)
{
  if (!xaintst) return 0;
  return xaintst->storeRecordNames;
}

XaintChar *
XaintStream_getRecordName(XaintStream *xaintst)
{
  __recordName *name;
  if (!xaintst) return 0;
  name = (__recordName*)xaintst->recordNames;
  while (name)
  {
    if (!name->next)
    {
      return name->recordName;
    }
    name = name->next;
  }
  return 0;
}

static XaintBool
__push_record_name(XaintStream *xaintst, XaintChar *name)
{
  __recordName *n_name, *l_record;
  if (!xaintst) return 0;
  n_name = (__recordName*)Xaint_malloc(sizeof(__recordName));
  if (!n_name) XaintParser_INTERNAL_oom(xaintst->xaintpr, 1);
  n_name->next = 0;

  if (name)
  {
    n_name->recordName = (XaintChar*)Xaint_malloc(strlen((char*)name)+1);
    if (!n_name->recordName)
    {
      Xaint_free(n_name);
      XaintParser_INTERNAL_oom(xaintst->xaintpr, 1);
    }
    strcpy((char*)n_name->recordName, (char*)name);
  }
  else
  {
    n_name->recordName = 0;  
  }


  if (!xaintst->recordNames)
  {
    xaintst->recordNames = n_name;  
  }
  else
  {
    l_record = (__recordName*)xaintst->recordNames;
    while (l_record)
    {
      if (!l_record->next)
      {
        l_record->next = n_name;
        break;
      }
      l_record = l_record->next;
    }
  }
  return 1;
}


static void
__free_record_name(XaintStream *xaintst)
{
  if (!xaintst) return;
  while (xaintst->recordNames)
  {
    __pop_record_name(xaintst);
  }
}

static void
__pop_record_name(XaintStream *xaintst)
{
  __recordName *p_record, *l_record;
  if (!xaintst) return;
  if (!xaintst->recordNames) return;

  if (!(((__recordName*)(xaintst->recordNames))->next))
  {
    if (((__recordName*)(xaintst->recordNames))->recordName) Xaint_free(((__recordName*)(xaintst->recordNames))->recordName);
    Xaint_free(xaintst->recordNames);
    xaintst->recordNames = 0;
    return;
  }

  p_record = (__recordName*)xaintst->recordNames;
  l_record = p_record->next;
  while (l_record)
  {
    if (!l_record->next)
    {
      if (l_record->recordName) Xaint_free(l_record->recordName);
      Xaint_free(l_record);
      p_record->next = 0;
      return;
    }
    p_record = l_record;
    l_record = l_record->next;
  }
}



void  
XaintStream_feedData(XaintStream *xaintst, 
                     XaintUnknown data, 
                     XaintNum size)
{
  XaintParser_feedData(xaintst->xaintpr, data, size);
}

static void
__send_pair_event(XaintStream *xaintst, 
                      XaintChar *value, 
                      XaintChar type)
{
  if (xaintst->pairEvent) 
  {
    xaintst->pairEvent(xaintst, xaintst->key, value, xaintst->ws, type);
    if (xaintst->storeRecordNames)
    {
      switch (type)
      {
      case XAINT_TYPE_SLIST:
        __push_record_name(xaintst, xaintst->key);
        break;
      case XAINT_TYPE_ELIST:
        __pop_record_name(xaintst);
        break;
      }
    }
  }
  if (xaintst->key) 
  {
    Xaint_free(xaintst->key);
    xaintst->key = 0;
  }
  if (xaintst->ws) 
  {
    Xaint_free(xaintst->ws);
    xaintst->ws = 0;
  }
}

static void 
__xaintprEvent(XaintParser *xaintpr, XaintChar *data, XaintChar type)
{
  XaintNum dsize;
  XaintStream *xaintst;
  xaintst = (XaintStream*)XaintParser_getUserData(xaintpr);
  
  if (data) dsize = strlen((char*)data)+1;
  else dsize = 0;
 
  switch (type)
  {
  case XAINT_TYPE_NAME:
    if (xaintst->key) __send_pair_event(xaintst, 0, type);
    xaintst->key = (XaintChar*)Xaint_malloc(dsize);
    if (!xaintst->key) XaintParser_INTERNAL_oom(xaintst->xaintpr, 1);
    memcpy(xaintst->key, data, dsize);
    break;
  case XAINT_TYPE_WS:
    if (xaintst->key)
    {
      xaintst->ws = (XaintChar*)Xaint_malloc(dsize);
      if (!xaintst->ws) XaintParser_INTERNAL_oom(xaintst->xaintpr, 1);
      memcpy(xaintst->ws, data, dsize);
    }
    else
    {
      if (xaintst->wsEvent) xaintst->wsEvent(xaintst, data);
    }
    break;
  case XAINT_TYPE_COMMENT:
    if (xaintst->key) __send_pair_event(xaintst, 0, XAINT_TYPE_NAME);
    if (xaintst->commentEvent) xaintst->commentEvent(xaintst, data);
    break;
  case XAINT_TYPE_PRAGMA:
    if (xaintst->key) __send_pair_event(xaintst, 0, XAINT_TYPE_NAME);
    if (xaintst->pragmaEvent) xaintst->pragmaEvent(xaintst, data);
    break;
  case XAINT_TYPE_ELIST:
    if (xaintst->key) __send_pair_event(xaintst, 0, XAINT_TYPE_NAME);
    __send_pair_event(xaintst, 0, XAINT_TYPE_ELIST);
    break;
  default:
    __send_pair_event(xaintst, data, type);
  }
}

static void
__xaintprDataRequest(XaintParser *xaintpr)
{
  XaintStream *xaintst;
  xaintst = (XaintStream*)XaintParser_getUserData(xaintpr);
  xaintst->dataRequest(xaintst);
}

XaintBool
XaintStream_parse(XaintStream *xaintst)
{
  XaintBool retval;
  if (!xaintst) return 0;

  XaintParser_setDataRequest(xaintst->xaintpr, __xaintprDataRequest);
  XaintParser_setEvent(xaintst->xaintpr, __xaintprEvent);
  retval = XaintParser_parse(xaintst->xaintpr);
  goto l__finished;

l__finished:
  if (xaintst->key) __send_pair_event(xaintst, 0, XAINT_TYPE_NAME); 
  return retval;
}

XaintStreamOnDelete
XaintStream_getOnDelete(XaintStream *xaintst)
{
  if (!xaintst) return 0;
  return xaintst->OnDelete;
}

void 
XaintStream_setOnDelete(XaintStream *xaintst, XaintStreamOnDelete OnDelete)
{
  if (!xaintst) return;
  xaintst->OnDelete = OnDelete;
}
void 
XaintStream_setClassInstance(XaintStream *xaintst, XaintUnknown ClassInstance)
{
  if (!xaintst) return;
  xaintst->ClassInstance = ClassInstance;
}
XaintUnknown 
XaintStream_getClassInstance(XaintStream *xaintst)
{
  if (!xaintst) return 0;
  return xaintst->ClassInstance;
}

static void 
__xaintStreamDataRequest(XaintStream *xaintst)
{
  if (xaintst->privData)
  {
    XaintStream_feedData(xaintst, xaintst->privData, strlen((char*)(xaintst->privData)));
    xaintst->privData = 0;
  }
  else
  {
    XaintStream_feedData(xaintst, 0, 0);
  }
}

XaintBool
XaintStream_parseString(XaintStream *xaintst, XaintChar *string)
{
  xaintst->privData = string;
  XaintStream_setDataRequest(xaintst, __xaintStreamDataRequest);
  return XaintStream_parse(xaintst);
}

